#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <glob.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <qapplication.h>
#include <qregexp.h>
#include <qmessagebox.h>
#include <qfiledialog.h>

#include "tcachedfile.h"
#include "aboutdialog.h"
#include "helpdialog.h"

//#define DUMMY_MODE
#define PROGRAM_NAME "UT2003 Cache Manager"

QStringList *mainForm::getCacheList() {{{
	QString ut2003pdir, cachedir, cacheini;
	static QStringList cache_list;
    char buffer[1024];
	FILE *cache_ini;
	bool in_list = false;

	ut2003pdir = getPersonalUTDir();
	cachedir   = ut2003pdir + "/Cache";
	cacheini   = cachedir   + "/cache.ini";

//	printf("Reading personal cache from %s\n", cachedir.ascii());

	if((cache_ini = fopen(cacheini.ascii(), "r")) == NULL) {
		perror("Unable to open cache.ini");
		emit qApp->quit();
		return NULL;
	}

	cache_list.clear();

	while(fgets(buffer, sizeof(buffer), cache_ini) != NULL) {
		QString strbuf = buffer;
		
		if(!in_list && !strcmp("[Cache]\n", buffer)) {
			in_list = true;
			continue;
		}
		
		strbuf.replace(QRegExp("\n*"), "");
		if(in_list && strbuf[0] != 0) {
//			printf("Cache: %s\n", strbuf.ascii());
			cache_list << strbuf;
		}
	}
	
	fclose(cache_ini);

	return &cache_list;
}}}
QString mainForm::getPersonalUTDir() {{{
	return QString(getenv("HOME")) + "/.ut2003";
}}}
void mainForm::fixCacheIni() {{{
	QStringList *cache = getCacheList();
	QStringList::Iterator it;
	QString file, cachedir;
	FILE *outfile;
	
	cachedir = getPersonalUTDir()+"/Cache/";

#ifndef DUMMY_MODE
	if((outfile = fopen((cachedir+"cache.ini").ascii(), "w")) == NULL) {
		perror("Unable to open cache.ini for writing");
		return;
	}
#else
	outfile = stdout;
#endif
	
	fprintf(outfile, "[Cache]\n");

	for(it = cache->begin(); it != cache->end(); it++) {
		file = cache2filename(*it);
		if(file[0] != 0) 
			fprintf(outfile, "%s\n", (*it).ascii());
	}

#ifndef DUMMY_MODE
	fclose(outfile);
#endif
	init();
}}}
QString & mainForm::cache2filename( QString &cache_entry ) {{{
	QString file = cache_entry;
	static QString retval;
	QString fullpath, cachedir;
	glob_t globbed;
	
	cachedir = getPersonalUTDir()+"/Cache/";

	file.replace(QRegExp("=.*$"), "");
	file += "*.uxx";

	fullpath = cachedir + file;
		
	glob(fullpath.ascii(), GLOB_NOSORT, NULL, &globbed);
	if(globbed.gl_pathc > 0) {
		retval = globbed.gl_pathv[0];
	} else {
		retval = "";
	}
	globfree(&globbed);

	return retval;
}}}
 
void mainForm::init() {{{
    struct dirent *direntry;
    DIR *dir;
    QString cachedir;
	QStringList *cache_list;

	cacheIconView->clear();
	moveIconView->clear();
	compressIconView->clear();
	
    cachedir   = getPersonalUTDir()+"/Cache";
	cache_list = getCacheList();

    if((dir = opendir(cachedir)) == NULL) {
		perror("Unable to open cache directory");
		emit qApp->quit();
		return;
    }

    while((direntry = readdir(dir))) {
		if(direntry->d_name[0] == '.') continue;

		QString name = direntry->d_name;
		QString fileid = name;

		fileid.replace(QRegExp("[^A-Z0-9a-z].*$"), "");
		QString realname = cache_list->grep(fileid).first();
		realname.replace(QRegExp("^.*="), "");

		if(realname[0] == 0) continue;
		
		new TCachedFile( name, realname, cacheIconView, NULL);
    }
 
    closedir(dir);

	checkButtons();
}}}

void mainForm::fileExit() {{{
    emit qApp->quit();
}}}
void mainForm::fileExtract() {{{
	QString tarball;
	
	tarball = QFileDialog::getOpenFileName(
		"",
		"Tarballs (*.tar.gz)",
		this,
		"open file dialog",
		"Choose a tarball to extract" );

	if(tarball == QString::null) return;

	QMessageBox::information(this, PROGRAM_NAME,
			"Extracting \n\n"+tarball+"\n\n"
			"to your personal UT2003 directory. See the\n"
			"output on the console for the list of files.");

	printf("Extracting %s:\n", tarball.ascii());
	system("tar -C "+getPersonalUTDir()+" -zxvf "+tarball);
}}}
void mainForm::helpHelp() {{{
	HelpDialog *help = new HelpDialog();
	help->exec();
}}}
void mainForm::helpAbout() {{{
	aboutDialog *about = new aboutDialog();
	about->exec();
}}}

QIconViewItem *mainForm::moveOne(bool skipcode) {{{ // returns a pointer to the next icon
	TCachedFile *current = (TCachedFile *)cacheIconView->currentItem();
	QIconViewItem *next;
	
	if(current == NULL) return NULL;

	if(CFT_code == current->getType()) {
		if(skipcode) return current->nextItem();

		switch( QMessageBox::warning( this, PROGRAM_NAME,
			"You are trying to move a code file from the cache:\n\n"
			+current->getRealName()+"\n\n"
			"This can lead to corruption of your UT2003 installation.\n"
			"It's usually better to find the mod you downloaded using\n"
			"Google (http://www.google.com/) and install it the usual\n"
			"way. You have been warned.\n\n"
			"Are you sure you want to continue moving this file?\n",
			QMessageBox::Yes,
			QMessageBox::Abort | QMessageBox::Escape | QMessageBox::Default )) {
		case QMessageBox::Yes:
			break;
		case QMessageBox::Abort:
			return NULL;
		}
	}

	new TCachedFile(
			current->getCachedName(),
			current->getRealName(),
			moveIconView);
	
	next = current->nextItem();
	delete current;

	checkButtons();
	
	return next;
}}}
QIconViewItem *mainForm::moveOne() {{{ // returns a pointer to the next icon
	return moveOne(false);
}}}
void mainForm::moveAll() {{{
	QIconViewItem *current = cacheIconView->firstItem();
	
	while(current != NULL) {
		cacheIconView->setCurrentItem(current);
		current = moveOne(true);
	}
}}}
bool mainForm::unMoveOne() {{{
	TCachedFile *current = (TCachedFile *)moveIconView->currentItem();

	if(current == NULL) return false;

	new TCachedFile(
			current->getCachedName(),
			current->getRealName(),
			cacheIconView);
	delete current;

	return true;
}}}
void mainForm::unMoveAll() {{{
	QIconViewItem *first;
	
	while((first = moveIconView->firstItem()) != NULL) {
		moveIconView->setCurrentItem(first);
		if(!unMoveOne()) return;
	}
}}}
void mainForm::move() {{{
	QString personaldir = getPersonalUTDir();
	QString destination;
	QString src, dst;
	TCachedFile *current, *next;
	
	if(moveIconView->count() == 0) {{{
		QMessageBox::information(this, PROGRAM_NAME,
				"First move the files you want to move out of the\n"
				"cache into the window called 'Move into UT2003'.\n"
				"Then press this button to move them into your\n"
				"UT2003.");
		return;
	}}}
	
	current = (TCachedFile *)moveIconView->firstItem();

	while(current != NULL) {
		switch(current->getType()) {
			case CFT_map: destination = "Maps"; break;
			case CFT_texture: destination = "Textures"; break;
			case CFT_sound: destination = "Sounds"; break;
			case CFT_music: destination = "Music"; break;
			case CFT_staticmesh: destination = "StaticMeshes"; break;
			case CFT_code: destination = "System"; break;
			default:
				QMessageBox::information(this, PROGRAM_NAME,
					"I don't know what to do with this file:\n\n"
					+current->getRealName()+"\n\n"
					"I'll skip this file. Please let the author know\n"
					"about this at meltdown@unrealtower.org. If you send\n"
					"an email, make sure you give the name of this file!\n");
				printf("Don't know what to do with %s - skipping\n", current->getRealName().ascii());
				current = (TCachedFile *)current->nextItem();
				continue;
				break;
		}

		printf("Moving %s to %s\n", current->getCachedName().ascii(), current->getRealName().ascii());
		mkdir((personaldir+"/"+destination).ascii(), 0755);

#ifndef DUMMY_MODE
		src = personaldir+"/Cache/"+current->getCachedName();
		dst = personaldir+"/"+destination+"/"+current->getRealName();
		if(link(src.ascii(), dst.ascii())) {
			QMessageBox::information(this, PROGRAM_NAME,
				"Unable to move \n"+src+" to\n"+dst+"\n\nError: "+
				strerror(errno));

			current = (TCachedFile *)current->nextItem();
			continue;
		}
		unlink(src);
#endif

		next = (TCachedFile *)current->nextItem();
		delete current;
		current = next;
	}

	fixCacheIni();
}}}

QIconViewItem * mainForm::compressOne( bool skipcode ) {{{ // returns a pointer to the next icon
	TCachedFile *current = (TCachedFile *)cacheIconView->currentItem();
	QIconViewItem *next;
	
	if(current == NULL) return NULL;

	if(CFT_code == current->getType()) {
		if(skipcode) return current->nextItem();

		switch( QMessageBox::warning( this, PROGRAM_NAME,
			"You are trying to move a code file from the cache:\n\n"
			+current->getRealName()+"\n\n"
			"This can lead to corruption of your UT2003 installation.\n"
			"It's usually better to find the mod you downloaded using\n"
			"Google (http://www.google.com/) and install it the usual\n"
			"way. You have been warned.\n\n"
			"Are you sure you want to continue moving this file?\n",
			QMessageBox::Yes,
			QMessageBox::Abort | QMessageBox::Escape | QMessageBox::Default )) {
		case QMessageBox::Yes:
			break;
		case QMessageBox::Abort:
			return NULL;
		}
	}

	new TCachedFile(
			current->getCachedName(),
			current->getRealName(),
			compressIconView);
	
	next = current->nextItem();
	delete current;

	checkButtons();
	return next;
}}}
QIconViewItem * mainForm::compressOne() {{{
	return compressOne(false);
}}}
void mainForm::compressAll() {{{
	QIconViewItem *current = cacheIconView->firstItem();
	
	while(current != NULL) {
		cacheIconView->setCurrentItem(current);
		current = compressOne(true);
	}

}}}
void mainForm::unCompressAll() {{{
	QIconViewItem *first;
	
	while((first = compressIconView->firstItem()) != NULL) {
		compressIconView->setCurrentItem(first);
		if(!unCompressOne()) return;
	}

}}}
bool mainForm::unCompressOne() {{{
	TCachedFile *current = (TCachedFile *)compressIconView->currentItem();

	if(current == NULL) return false;

	new TCachedFile(
			current->getCachedName(),
			current->getRealName(),
			cacheIconView);
	delete current;

	return true;

}}}
void mainForm::compess() {{{
	QString personaldir = getPersonalUTDir();
	QString destination;
	QString src, dst, cpcmd;
	QString tarname;
	char tardir[] = "/tmp/cachecompress-XXXXXX";
	TCachedFile *current, *next;
	bool in_error = false;
	int sysstatus;

	if(compressIconView->count() == 0) {
		QMessageBox::information(this, PROGRAM_NAME,
				"First move the files you want to compress into\n"
				"the window called 'Compress into .tar.gz'. Then\n"
				"press this button to compress them into a tarball.");
		return;
	}

	current = (TCachedFile *)compressIconView->firstItem();

	if(compressIconView->count() == 1) {
		tarname = current->getRealName();
		tarname.replace(QRegExp("\\..*$"), ".tar.gz");
	} else {
		tarname = "tarball.tar.gz";
	}
	tarname = QFileDialog::getSaveFileName(
			tarname,
			"Tarballs (*.tar.gz)",
			this,
			"save file dialog"
			"How and where shall I save your tarball?" );
	if(tarname == QString::null) return;
	
	if(mkdtemp(tardir) == NULL) {{{
		QMessageBox::critical(this, PROGRAM_NAME,
					"Unable to create temporary directory:\n\n"
					+QString(tardir)+"\n\n"
					"Please check the permissions in /tmp\n");
		perror("Unable to create temporary directory");
		return;
	}}}

	while(current != NULL) {
		switch(current->getType()) {{{
			case CFT_map: destination = "Maps"; break;
			case CFT_texture: destination = "Textures"; break;
			case CFT_sound: destination = "Sounds"; break;
			case CFT_music: destination = "Music"; break;
			case CFT_code: destination = "System"; break;
			default:
				QMessageBox::information(this, PROGRAM_NAME,
					"I don't know what to do with this file:\n\n"
					+current->getRealName()+"\n\n"
					"I'll skip this file. Please let the author know\n"
					"about this at meltdown@unrealtower.org. If you send\n"
					"an email, make sure you give the name of this file!\n");
				printf("Don't know what to do with %s - skipping\n", current->getRealName().ascii());
				current = (TCachedFile *)current->nextItem();
				continue;
				break;
		}}}

		printf("Compressing %s to %s\n", current->getCachedName().ascii(), current->getRealName().ascii());

#ifndef DUMMY_MODE
		destination = QString(tardir)+"/"+destination;
		mkdir(destination, 0755);
		src   = personaldir+"/Cache/"+current->getCachedName();
		dst   = destination+"/"+current->getRealName();
		cpcmd = "cp '"+src+"' '"+dst+"'";
		
		sysstatus = system(cpcmd);
		if(sysstatus == -1 || WEXITSTATUS(sysstatus) > 0) {
			perror("Unable to copy "+src+" to "+dst);
			QMessageBox::warning(this, PROGRAM_NAME,
					"Unable to move \n\n"+src+"\n\nto\n\n"+dst);
			current  = (TCachedFile *)current->nextItem();
			in_error = true;
			continue;
		}
#endif

		next = (TCachedFile *)current->nextItem();
		delete current;
		current = next;
	}

	if(in_error) {
		QMessageBox::information(this, PROGRAM_NAME,
				"There were errors, so I won't create the tarball");
	} else {
		char cwd[2048];
		
		getcwd(cwd, sizeof(cwd));
		QMessageBox::information(this, PROGRAM_NAME,
			"Saving as "+tarname);
		chdir(tardir);
		system("tar -zcf "+tarname+" *");
		chdir(cwd);
	}

	system(QString("rm -rf ")+tardir);

	fixCacheIni();
}}}

void mainForm::deleteCache() {{{
	QStringList *cache = getCacheList();
	QString file;
	QStringList::Iterator it;

	unCompressAll();
	unMoveAll();
	
	if(cacheIconView->count() == 0) {
		QMessageBox::information(this, PROGRAM_NAME,
				"There are no files in the cache, so there is\n"
				"nothing to do for this sweet, small button\n");
		return;
	}
	
	switch( QMessageBox::warning( this, PROGRAM_NAME,
		"Are you sure you want to clean the entire cache?\n"
		"This operation can't be reversed!\n",
		QMessageBox::Yes | QMessageBox::Default,
		QMessageBox::Abort | QMessageBox::Escape) ) {
	case QMessageBox::Yes:
		break;
	default:
		return;
	}

	for(it = cache->begin(); it != cache->end(); it++) {
		file = cache2filename(*it);
#ifndef DUMMY_MODE
		if(unlink(file)) perror("Unable to delete file");
#else
		printf("Removing %s\n", file.ascii());
#endif
	}

	fixCacheIni();
}}}
void mainForm::deleteCode() {{{
	QStringList *cache = getCacheList();
	QString file;
	QStringList::Iterator it;

	unCompressAll();
	unMoveAll();
	
	if(cache->grep("\\.u$").count() == 0) {
		QMessageBox::information(this, PROGRAM_NAME,
				"There are no *.u files in the cache, so there\n"
				"is nothing to do for this sweet, small button\n");
		return;
	}

	switch( QMessageBox::warning( this, PROGRAM_NAME,
		"Are you sure you want to remove all code from the cache?\n"
		"This operation can't be reversed!\n",
		QMessageBox::Yes | QMessageBox::Default,
		QMessageBox::Abort | QMessageBox::Escape) ) {
	case QMessageBox::Yes:
		break;
	default:
		return;
	}

	for(it = cache->begin(); it != cache->end(); it++) {
		if((*it).contains(QRegExp(".u$", false))) {
			file = cache2filename(*it);
#ifndef DUMMY_MODE
			if(unlink(file)) perror("Unable to delete file");
#else
			printf("Removing %s\n", file.ascii());
#endif
		}
	}

	fixCacheIni();
}}}

void mainForm::checkButtons() {{{
	/*
	bool delBtnsEnabled;

	delBtnsEnabled = (
			moveIconView->count() == 0
			&& compressIconView->count() == 0
			&& cacheIconView->count() > 0);

	deleteCacheButton->setEnabled(delBtnsEnabled);
	deleteCodeButton->setEnabled(delBtnsEnabled);
	*/
}}}

